function [ approx ] = MG_V(A, f, I, smoother, solver_parameters, x_init)
% Multigrid solver, one V-cycle iteration
%
%   Jan Papez, April 2025
%
% INPUT:
% A... problem matrix on all levels (structure A{1}, A{2}, ..., A{J})
% f... rhs on the finest level J
% I... interpolation (prolongation) matrix (structure I{1}, I{2}, ..., I{J-1})
% smoother (as a function handle)
% solver_parameters... number of pre- and post-smoothing steps
% x_init... (optional) initial approximation
%
% OUTPUT: 
% approx... computed approximation

% number of levels
J = length(A);

if nargin < 5 || isempty(x_init)
    approx = Vcycle(A, f, J, I, smoother, solver_parameters);
else
    approx = Vcycle(A, f - A{J}*x_init, J, I, smoother, solver_parameters);
    approx = approx + x_init;
end

end


function [d] = Vcycle(A, r, j, I, smoother, solver_parameters)
    if j ~= 1
        prolong_matrix = I{j-1};
        
        % presmoothing
        d = smoother(A{j}, r, zeros(size(r)), solver_parameters{1});
        % coarse-space correction
        corr = Vcycle(A, 1/4* prolong_matrix'*(r - A{j}*d), j-1, I, smoother, solver_parameters);
        %postsmoothing
        d = smoother(A{j}, r, d + prolong_matrix*corr, solver_parameters{2});
    else
        %direct solve on the coarsest level
        d = A{1}\r;
    end
    
end



