%% script for running MG solver with varying parameters, for the Multilevel 
% Methods course
%
% Jan Papez, April 2025

% 1D geometric multigrid on the interval (0,1) with homogeneous Dirichlet
% boundary condition and finite-differences discretization

clear, clc;

%% parameters
J = 6;
smoother = @gs;
presm = 3;
postsm = 3;
n_J = 1024;
maxiter = 20;

h = 1/n_J;

% choice of the test case (rhs)
% uex = @(x) sin(pi*x);
% f = @(x) pi^2*sin(pi*x);

uex = @(x) x.*(x-1);
f = @(x) -2*ones(size(x));

%% assembly of the matrix and the matrices on the coarsest levels
% including the iterpolation matrices, with coarsening factor = 2;
% restriction is given by 1/2 I{j}'

stencil = [1/2 1 1/2]';
I = struct([]);

e = ones(n_J-1,1);
A{J} = 1/(h^2) * spdiags([-e 2*e -e], -1:1, n_J-1, n_J-1);

tmp = linspace(0,1,n_J+1);
F_J = f(tmp(2:end-1)');

h_j = h;
n_j = n_J;

for j = J-1:-1:1
    h_j = 2*h_j;
    n_j = n_j/2;
    
    e = ones(n_j-1,1);
    A{j} = 1/(h_j^2) * spdiags([-e 2*e -e], -1:1, n_j-1, n_j-1);  

    %interpolation matrix is assembled as a sparse matrix
    I{j} = sparse(size(A{j+1},2),size(A{j},2));

    for i_c = 1:size(A{j},2)    % iterate over all coarse nodes
        I{j}( (2*i_c + [-1:1]) , i_c) = stencil;
    end

end

%% computing the exact solution
if size(A{J},2) < 1e4
    u = A{J}\F_J;
else % if the problem is too big, overiterate by MG solver
    u = zeros(size(A{J},2),1);
    for i = 1:30
        u = MG_V(A, F_J, I, @gs, {10,10}, u);
    end
end


%% run V-cycle
ui = zeros(size(A{J},2),1);
e0_norm = norm(u-ui);

for i = 1:maxiter
    ui = MG_V(A, F_J, I, smoother, {presm,postsm}, ui);
    ei_norm(i) = norm(u-ui);
end

%% plot the convergence
rel_error_norm = [1,ei_norm/e0_norm];

figure(1);
semilogy(0:length(ei_norm),rel_error_norm, '-o','LineWidth',3);
title('MG convergence')

% evaluate the contraction factor
contr_faktor = rel_error_norm(2:end)./rel_error_norm(1:end-1);
figure(2);
plot(contr_faktor, '-o','LineWidth',3);
title('error contraction in the iterations')


