%% script for running MG solver with varying parameters, for the Multilevel 
% Methods course
%
% Jan Papez, April 2025

% 2D geometric multigrid on the domain (0,1)x(0,1) with homogeneous 
% Dirichlet boundary condition and finite-differences discretization

clear, clc;

%% parameters
J = 5;
smoother = @jacobi;
presm = 1;
postsm = 1;
m_J = 1024;
n_J = m_J;
maxiter = 20;

hx = 1/m_J;
hy = 1/n_J;

% choice of the test case (rhs)
uex = @(x,y) sin(pi*x).*sin(pi*y);
f = @(x,y) 4*pi^2*sin(pi*x).*sin(pi*y);

% uex = @(x,y) (x).*(x-1).*(y).*(y-1);
% f = @(x,y) -2*(x.^2-x+y.^2-y);

% exponent = @(x,y) exp( -100*(x-0.5).^2 - 100*(y-117/1000).^2 );
% uex = @(x,y) x.*(x-1).*y.*(y-1).* exponent(x,y);
% f = @(x,y) - 1/25 * exponent(x,y) .* ...
%     (1000000 * x.^4 .* (y-1) .* y - 2000000 * x.^3 .* (y-1) .* y + ...
%     x.^2 .* (1000000 .* y.^4 - 1234000 * y.^3 + 1447689 * y.^2 - 1221349 * y - 1120) + ...
%     x .* (-1000000 .* y.^4 + 1234000 * y.^3 - 447689 * y.^2 + 221349 * y + 1120)- ...
%     4950 * (y-1) .* y);


%% assembly of the matrix and the matrices on the coarsest levels
% including the iterpolation matrices, with coarsening factor = 2;
% restriction is given by 1/4 I{j}'

stencil = [1/4 1/2 1/4; 1/2 1 1/2; 1/4 1/2 1/4];
stencil = stencil(:);
I = struct([]);

[X,Y] = meshgrid(1/m_J:1/m_J:1-1/m_J,1/n_J:1/n_J:1-1/n_J);

F_J = f(X(:),Y(:));

e = ones(m_J-1,1);
B = 1/(hy^2) * spdiags([-e 2*(hx^2+hy^2)/(hx^2)*e -e], -1:1, m_J-1, m_J-1);

e = ones(n_J-1,1);    
tmp = spdiags([e zeros(size(e)) e], -1:1, n_J-1, n_J-1);
A{J} = kron(speye(n_J-1), B) + ...
    kron(tmp, -1/(hx^2) * speye(m_J-1));    

hy_j = hy; hx_j = hx;
m_j = m_J; n_j = n_J;

for j = J-1:-1:1
    hy_j = 2*hy_j; hx_j = 2*hx_j;
    m_j = m_j/2; n_j = n_j/2;
    
    e = ones(m_j-1,1);
    B = 1/(hy_j^2) * spdiags([-e 2*(hx_j^2+hy_j^2)/(hx_j^2)*e -e], -1:1, m_j-1, m_j-1);
    
    e = ones(n_j-1,1);    
    tmp = spdiags([e e], [-1, 1], n_j-1, n_j-1);
    A{j} = kron(speye(n_j-1), B) + ...
        kron(tmp, -1/(hx_j^2) * speye(m_j-1));  
    
    %interpolation matrix is assembled as a sparse matrix
    I_slindex = zeros(size(A{j},2) * length(stencil),1);
    I_radindex = I_slindex;
    I_hodnoty = I_slindex;
    index = 0;

    for i_c = 1:size(A{j},2)    % iterate over all coarse nodes
        column = floor(i_c/ (m_j-1)); row = mod(i_c, (m_j-1));
        if row == 0, column = column-1; row = m_j-1; end
        i_f = (2*column + 1) * (2*m_j - 1) + 2*row;  % index of this node on the fine level
        indexy_uzlu_kolem = [[i_f-1,i_f,i_f+1]-(2*m_j-1) , [i_f-1,i_f,i_f+1], [i_f-1,i_f,i_f+1] + (2*m_j-1)];
        
        I_slindex(index+1:index+length(stencil)) = i_c;
        I_radindex(index+1:index+length(stencil)) = indexy_uzlu_kolem;
        I_hodnoty(index+1:index+length(stencil)) = stencil;
        index = index + length(stencil);
    end
    I{j} = sparse(I_radindex,I_slindex,I_hodnoty,size(A{j+1},2),size(A{j},2));

end


%% computing the exact solution
if size(A{J},2) < 1e4
    u = A{J}\F_J;
else % if the problem is too big, overiterate by MG solver
    u = zeros(size(A{J},2),1);
    for i = 1:30
        u = MG_V(A, F_J, I, @gs, {10,10}, u);
    end
end


%% run V-cycle
ui = zeros(size(A{J},2),1);
e0_norm = norm(u-ui);

for i = 1:maxiter
    ui = MG_V(A, F_J, I, smoother, {presm,postsm}, ui);
    ei_norm(i) = norm(u-ui);
end

%% plot the convergence
rel_error_norm = [1,ei_norm/e0_norm];

figure(1);
semilogy(0:length(ei_norm),rel_error_norm, '-o','LineWidth',3);
title('MG convergence')

% evaluate the contraction factor
contr_faktor = rel_error_norm(2:end)./rel_error_norm(1:end-1);
figure(2);
plot(contr_faktor, '-o','LineWidth',3);
title('error contraction in the iterations')


